/***********************************************
// Test file for RFM69 RX TX on ARDURF1 or ARDUFRF1s or PC2ARDURF1
// Uses LowPowerLabs RFM69 lib as of 12/11/2014
// or RADIOHEAD RF69 lib ver 1.37
// based on LowPowerLab RFM69 examples and Radiohead RF69 examples
//
// Note: it looks like both rf69 libraries default to using arduino Serial port to dump the radios register values

// USAGE:
// Comment out or in the defines inside the RADIO_EN define
// to select:
// A. which rfm69 library to use LowPowerLabs RFM69 or RadioHead RF69
// B. which ardurf board to use it with: ARDURF1, ARDURF1s or PC2ARDURF1
// C. whether or not the ardurf board is primarily a SENDER (aka Client, Node) or a LISTENER (aka Server, Gateway)
//    and what nodeids the board uses.


// BROADCASTMODE   has  been tested on both LOWPOWERLABS RFM69 lib and Radiohead RF69 lib
// PROMISCUOUSMODE has  been tested on both LOWPOWERLABS RFM69 lib and Radiohead RF69 lib
// The tx and rx code sends and receives a data packet contained in a C struct called Payload.

// You can comment out the SERIAL_EN, VCCMONITOR_EN, WATCHDOG_EN and SLEEP_EN defines
// to save memory, the radio will still work.
//
// If SERIAL_EN is defined it sets the BAUDRATE, allows debug messages to be sent to Serial port
// and a Menu will be displayed at startup allowing the user to set and query the operating status used by the code and radio.
// press h or ? at antime to display the menu.
// Menu options include enabling and disabling Broadcast and Promiscuous Mode (not yet working with Radiohead lib)
// Toggling the RX and or TX off just prevent the radioListen and or radioSend code from running.

// If VCCMONITOR_EN is defined the code will report the boards Vcc, the Ardrfs Vbatt and IChrg values
// Not sure if the showArduRFIChrg() and the showArduRFVBatt()
// are processing the ArduRF batt and Ichrg info on the A6 and A7 pins correctly
// Not sure if the volts set within the VCCMONITOR_EN defines are set
// correctly for each of the 5v and 3.3 volt ArduRfs

// If WATCHDOG_EN is defined and SLEEP_EN is NOT defined then a 4 second watchdog timer is setup to reboot the board
// if something hangs up the processor for longer than 4 seconds.
// it has been tested with the UNO and Mega 2560 processors.

// If WATCHDOG_EN is defined SLEEP_EN can be defined
//   If SLEEP_EN is defined the Watchdogtimer will be set up to trigger every 8 seconds and not reboot the board.
//   This can be used to to wake up a board from Sleeping

// If SLEEP_EN is defined it defines USE_LOWPOWER_LIB
// which uses a library called: LowPower lib.
// See LowPower.h - it may not work with the 1284p processor.
// there is other code in this file that could be used instead of using the LowPower lib
// but even with USE_LOWPOWER_LIB undefined
// the compiler reports: multiple definition of `__vector_6
// aka ISR(WDT_vect) so for now must use the LowPower lib to sleep
// If SLEEP_EN is undefined still must comment out the #include <LowPower.h>
// May need to upgrade to a newer arduino ide to resolve this compiler error.
//
//
// LOW POWER SLEEP INFO
//    A packet will only be received if the radio is listening.
//    If it's sleeping it will not "hear" anything. (receiverDone returns false).
//    And I am assuming it cant send anything too. send and sendwithretry fail.
//    So incoming packets while be dropped while sleeping and no data will can be sent.
//    I would like to wake up from sleep via the watchdog and or when a radio received a packet.
//

// Added code to lowpower lab receiver see runradioListen() to send a reply message back 
// to sender after ack just like radiohead receiver


// Things to do:
// A. Need to fix: A listener should not try to send to itself (see SendDelay)

// Obervations and Questions:
// FreeSram:
// With nothing defined (RADIO_EN, SERIAL_EN, WATCHDOG_EN, VCCMONITOR_EN ALL NOT DEFINED)
// the system (arduino uno) reports 1567 bytes of feesram.
// With Everthing defined (RADIO_EN, SERIAL_EN, WATCHDOG_EN, VCCMONITOR_EN ALL DEFINED) using the LowPowerLabs RF69M lib 
// the system (arduino uno) reports 1056 bytes of feesram.
// With Everthing defined (RADIO_EN, SERIAL_EN, WATCHDOG_EN, VCCMONITOR_EN ALL DEFINED) using the Radiohead RF69 lib
// the system (arduino uno) reports 557 bytes of feesram.

// Battery: 
// Plugging the battery in and out is a little difficult, sometimes I have to wiggle the connector to get it to work.
// I have had one battery charging via the usb port for 24 hours and no issues.
// I have had an ARDURF1s running on a fully charged battery sending HELLO WORLD once every 5 seconds
// for 80 hours before the battery shutdown. I expect much longer run time when SLEEP_EN is defined.
// 
// Uno Reset Button:
// it does not reset the radio, have to remove power to reset the radio if it locks up.
//
// Low Power Labs RFM69 lib:
// A. the lib needs a way to specify the modemconfig to use so a define can be used to set it.
// A. the init routine and possibly the cansend routine need timeouts so if their while loops hang they can exit
//    and return false.
// B. all set functions should return a boolean indicating success or failure.
// C. all set functions should also have a get function so the app code can query the radios current configurations.
// D. needs functions to set and get a radios MODEMCONFIG parameter.
// E. needs function to get a radios IDLE and Standby mode status.
// F. it issues Serial.print command within certain functions, such as:
//    readAllRegs, readTemperature 
//
// Radiohead RF69 lib:
// A. the lib needs an easier way to specify the modemconfig to use so a define can be used to set it.
// B. all set functions should return a boolean indicating success or failure.
// C. all set functions should also have a get function so the app code can query the radios current configurations.
// D. needs function to get a radios MODEMCONFIG parameter.
// E. needs functions to get a radios IDLE and Standby modes status.
// F. it issues Serial.print command within certain functions, such as:
//    printRegisters
//

//**********************************************
On Arduino Uno AVR328p the spi connections are:
RFM69W To Arduino

    NSS  to  Pin 10   (aka SPI Chip Select)
    MOSI to  Pin 11
    MISO to  Pin 12
    SCK  to  Pin 13
    GND  to  Ground
    3.3V to  the 3.3V header
    DI00 to  Pin 2 (interrupt)
//**********************************************
************************************************/

const char AppVer[9] = "V1.19";

//****************** for radio
#define RADIO_EN
  #ifdef RADIO_EN
    
    //#define USE_LOWPOWERLABS_RFM69_LIB
    #define USE_RADIOHEAD_RF69_LIB
    
    //#define RADIO_SENDER          // uncomment this if this radio primarily acts as a Node or CLIENT
    #define RADIO_LISTENER        // uncomment this if this radio primarily acts as a GATEWAY or SERVER
    
    //#define ARDURF1
    #define PC2ARDURF1    // usually the LISTENER
    //#define ARDURF1S    
    
    #if not defined USE_LOWPOWERLABS_RFM69_LIB && not defined USE_RADIOHEAD_RF69_LIB
      #error "Must define either USE_LOWPOWERLABS_RFM69_LIB or USE_RADIOHEAD_RF69_LIB"
    #endif    
    
    #if defined USE_LOWPOWERLABS_RFM69_LIB && defined USE_RADIOHEAD_RF69_LIB
      #error "Must define either USE_LOWPOWERLABS_RFM69_LIB or USE_RADIOHEAD_RF69_LIB not both"
    #endif    
    
    #if not defined RADIO_SENDER && not defined RADIO_LISTENER
      #error "Must define either RADIO_SENDER or RADIO_LISTENER"
    #endif   
    
    #if defined RADIO_SENDER && defined RADIO_LISTENER
      #error "Must define either RADIO_SENDER or RADIO_LISTENER not both"
    #endif 
    
    #if not defined ARDURF1 && not defined PC2ARDURF1 && not defined ARDURF1S
      #error "Must define either ARDURF1 or PC2ARDURF1 or ARDURF1S"
    #endif 
    
    #if defined ARDURF1 && defined PC2ARDURF1 && defined ARDURF1S
      #error "Must define either ARDURF1 or PC2ARDURF1 or ARDURF1S not all three"
    #endif
    
    #if defined ARDURF1 && defined PC2ARDURF1 
      #error "Must define either ARDURF1 or PC2ARDURF1 not both"
    #endif
    
    #if defined ARDURF1 && defined ARDURF1S 
      #error "Must define either ARDURF1 or ARDURF1S not both"
    #endif
    
    #if defined PC2ARDURF1 && defined ARDURF1S
      #error "Must define either PC2ARDURF1 or ARDURF1S not both"
    #endif

    #ifdef RADIO_LISTENER
      #ifdef RADIO_SENDER
        #undef RADIO_SENDER
      #endif
    #endif

    #ifdef RADIO_SENDER
      #ifdef RADIO_LISTENER
        #undef RADIO_LISTENER
      #endif
    #endif


    #ifdef USE_LOWPOWERLABS_RFM69_LIB
      #ifdef USE_RADIOHEAD_RF69_LIB
        #undef USE_RADIOHEAD_RF69_LIB
      #endif
    #endif

    #ifdef USE_RADIOHEAD_RF69_LIB
      #ifdef USE_LOWPOWERLABS_RFM69_LIB
        #undef USE_LOWPOWERLABS_RFM69_LIB
      #endif
    #endif

    #ifdef ARDURF1
      #ifdef PC2ARDURF1
        #undef PC2ARDURF1
      #endif
      #ifdef ARDURF1S
        #undef ARDURF1S
      #endif
    #endif

    #ifdef PC2ARDURF1
      #ifdef ARDURF1
        #undef ARDURF1
      #endif
      #ifdef ARDURF1S
        #undef ARDURF1S
      #endif
    #endif

    #ifdef ARDURF1S
      #ifdef ARDURF1
        #undef ARDURF1
      #endif
      #ifdef PC2ARDURF1
        #undef PC2ARDURF1
      #endif
    #endif

    #define SENDDELAY 5000  // Time in ms between updates used when SLEEP_EN is not defined
    unsigned long      previousSendMillis   = 0;              // will store last time LED was updated
    long               sendinterval         = SENDDELAY;    // interval at which to blink (milliseconds)
    #define ACK_TIME   1000 // max # of ms to wait for an ack
    boolean            toggleDestinationID  = false;

    #if defined RADIO_EN && defined USE_LOWPOWERLABS_RFM69_LIB
      #include <RFM69.h>    //get it here: https://github.com/LowPowerLab/RFM69
      #include <SPI.h>
      #define NETWORKID               100  // the same on all nodes that talk to each other
      #define BOGUS_NODEID            55   // send to some unknown node

      #ifdef ARDURF1
        #define NODE_ADDRESS           2    // tx nodeid, unique for each node on same network
        #define GATEWAY_ADDRESS        1    // rx nodeid, unique for each node on same network
      #else
        #ifdef ARDURF1S
          #define NODE_ADDRESS         3    // tx nodeid, unique for each node on same network
          #define GATEWAY_ADDRESS      1    // rx nodeid, unique for each node on same network
        #else
          #ifdef PC2ARDURF1
            #define NODE_ADDRESS       4    // tx nodeid, PC2ARDURF1 usually acts as GATEWAY or SERVER
            #define GATEWAY_ADDRESS    1    // rx nodeid, unique for each node on same network
          #endif
        #endif
      #endif

      #ifdef RADIO_LISTENER
        #define MYNODEID         GATEWAY_ADDRESS
        #define SENDTONODEID     4      // chose a senders nodeid to send to, just dont choose your nodeid, i.e. dont send to self
      #else                             // define broadcast mode above to send to all other nodes on the network
        #ifdef RADIO_SENDER
          #define MYNODEID       NODE_ADDRESS
          #define SENDTONODEID   GATEWAY_ADDRESS      // set this to the nodeid of the listener
        #endif
      #endif

      //Match frequency to the hardware version of the radio :
      //#define FREQUENCY     RF69_433MHZ
      //#define FREQUENCY     RF69_868MHZ
      #define FREQUENCY     RF69_915MHZ
      #define ENCRYPTKEY    "crosspollination" //exactly the same 16 characters/bytes on all nodes!

      //#define IS_RFM69HW   //NOTE: uncomment this ONLY for RFM69HW or RFM69HCW
      #ifdef IS_RFM69HW
        #define TXPOWERSETTING  "High"
      #else
        #define TXPOWERSETTING  "Normal"
      #endif

      #define MODEMCONFIG       "RFM69DEFAULT"

      //#define LED           13 No can use pin 13 (led on uno because default spi uses pin 13
      RFM69 radio;
      #define MAXPAYLOADTEXTLEN      30
      typedef struct 
      {		
        int                 nodeId; //store this nodeId
        unsigned long       uptime; //uptime in ms
        float               vcc;    //board vcc
        float               vbat;   //board vbat
        int                 ichrg;  //board icharge in ma
        char  dataStr[MAXPAYLOADTEXTLEN];   // total sizeof struct must be <= RF69_MAX_DATA_LEN
      } Payload;
      Payload theData;
      
      boolean  promiscuousMode = false; //set to 'true' to sniff all packets on the same network
      boolean  encryptMode     = false; //set to 'true' to encrypt all packets on the same network
      boolean  rxMode = false;
      boolean  txMode = false;
      boolean  broadcastMode = false;

      /********************
      SPIFlash flash(FLASH_SS, 0xEF30); //EF30 for 4mbit  Windbond chip (W25X40CL)
      ********************/
    #endif
    //***************** end use lowpowerlabs rfm69 library

    #if defined RADIO_EN && defined USE_RADIOHEAD_RF69_LIB
      // showing how to create a simple addressed, reliable messaging server
      // with the RHReliableDatagram class, using the RH_RF69 driver to control a RF69 radio.
      // It is based on examples: rf69_reliable_datagram_server.pde and  rf69_reliable_datagram_client
      #include <RHReliableDatagram.h>
      #include <RH_RF69.h>
      #include <SPI.h>

      //#define CLIENT_ADDRESS 1
      //#define SERVER_ADDRESS 2
      #define BOGUS_NODEID            55   // send to some unknown node

      // syncwords act as a network id has to be the same as all nodes
      uint8_t syncwords[] = { 0x2d, 0xd4 };

      // The encryption key has to be the same as the one in the server
      uint8_t rh_ekey[] = { 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08};

      #ifdef ARDURF1
        #define CLIENT_ADDRESS        2    // tx nodeid, unique for each node on same network
        #define SERVER_ADDRESS        1    // rx nodeid, unique for each node on same network
      #else
        #ifdef ARDURF1S
          #define CLIENT_ADDRESS      3    // tx nodeid, unique for each node on same network
          #define SERVER_ADDRESS      1    // rx nodeid, unique for each node on same network
        #else
          #ifdef PC2ARDURF1
            #define CLIENT_ADDRESS    4    // tx nodeid, PC2ARDURF1 usually acts as GATEWAY or SERVER
            #define SERVER_ADDRESS    1    // rx nodeid, unique for each node on same network
          #endif
        #endif
      #endif

      #ifdef RADIO_LISTENER
        #define MYNODEID         SERVER_ADDRESS
        #define SENDTONODEID     4      // chose a senders nodeid to send to, just don chose your nodeid, i.e. dont send to self
      #else                             // define broadcast mode above to send to all other nodes on the network
        #ifdef RADIO_SENDER
          #define MYNODEID       CLIENT_ADDRESS
          #define SENDTONODEID   SERVER_ADDRESS      // set this to the nodeid of the listener
        #endif
      #endif

      // Singleton instance of the radio driver
      RH_RF69 driver;                // get it here: http://www.airspayce.com/mikem/arduino/RadioHead/


      // Class to manage message delivery and receipt, using the driver declared above
      #ifdef RADIO_SENDER
        RHReliableDatagram manager(driver, CLIENT_ADDRESS);
      #else
        #ifdef RADIO_LISTENER
          RHReliableDatagram manager(driver, SERVER_ADDRESS);
        #endif
      #endif

    //Match frequency to the hardware version of the radio.
      //#define FREQUENCY     433.0
      //#define FREQUENCY     868.0
      #define   FREQUENCY     915.0

      //#define IS_RFM69HW   //NOTE: uncomment this ONLY for RFM69HW or RFM69HCW
      #ifdef IS_RFM69HW
        #define TXPOWERSETTING  18  // ONLY ONLY if radio is  RFM69HW or RFM69HCW
      #else
        #define TXPOWERSETTING  13  // same as the default
      #endif
      #define PREAMBLELEN       4
         
      // this is in effect an index into an array, see RH_69.cpp PROGMEM static const RH_RF69::ModemConfig MODEM_CONFIG_TABLE[]
      #define MODEMCONFIG       GFSK_Rb250Fd250 
       
      #define MAXPAYLOADTEXTLEN      30
      typedef struct 
      {		
        int                 nodeId; //store this nodeId
        unsigned long       uptime; //uptime in ms
        float               vcc;    //board vcc
        float               vbat;   //board vbat
        int                 ichrg;  //board icharge in ma
        char  dataStr[MAXPAYLOADTEXTLEN];   // total sizeof struct must be <= RH_RF69_MAX_MESSAGE_LEN
      } Payload;
      Payload theData;

      boolean  radioPresent    = false; // see initRHmanager
      boolean  promiscuousMode = false; //set to 'true' to sniff all packets on the same network
      boolean  encryptMode     = false; //set to 'true' to encrypt all packets on the same network
      boolean  rxMode          = false;
      boolean  txMode          = false;
      boolean  broadcastMode = false;

    #endif
    //***************** end use radiohead rf69 library
#endif
//*********** end RADIO_EN **********************


#define SERIAL_BAUD   115200
  #if defined RADIO_EN && defined USE_LOWPOWERLABS_RFM69_LIB
    const int  MAXARRAYSIZE   = RF69_MAX_DATA_LEN;             // 61 chars
  #else
    #if defined RADIO_EN && defined USE_RADIOHEAD_RF69_LIB
      const int  MAXARRAYSIZE   = RH_RF69_MAX_MESSAGE_LEN;     // 60 chars
    #else
      const int  MAXARRAYSIZE   = 64;     // 63 chars plus 1 for \0 terminator
    #endif
  #endif

const int  MAXNUMCHARS    = MAXARRAYSIZE - 1;  // leave room for null terminator
char  tempoutString1[MAXARRAYSIZE];  // for formatting data

#define SERIAL_EN      //uncomment this line to enable serial IO debug messages, leave out if you want low power
#ifdef SERIAL_EN
  #define DEBUG(input)   {Serial.print(input); delay(1);}
  #define DEBUGln(input) {Serial.println(input); delay(1);}
  #define DEBUGflush()   {Serial.flush(); delay(1);}
  #define DEBUGHex(input1,input2) {Serial.println(input1,input2); delay(1);}
  #define DEBUGHexln(input1,input2) {Serial.println(input1,input2); delay(1);}
#else
  #define DEBUG(input);
  #define DEBUGln(input);
  #define DEBUGflush();
  #define DEBUGHex(input1,input2);
  #define DEBUGHexln(input1,input2);
#endif
//*********** end SERIAL_EN **********************

// for watchdog
/************************************
 Note, after a System Reset the WDT is still enabled,
 also the configuration will have reverted to default, i.e. 16ms time-out,
 If the user doesn't re-configure or disable the WDT before that time-out
 you can end up with the micro in a continuous System Reset loop.
 The following will disable the WDT before user code is executed:

// @brief Disables the watchdog at startup, otherwise after a timeout reset
// it will continue to timeout with it's default value (<50mS).
// The get_mcuser() function is configured (by the __attribute__'s) to be
// automatically executed before any of your normal user code will be
// executed (e.g. loop() and setup()).
// It has to do with memory layout, have a look at linker scripts for more info.
// It can be placed anywhere in your sketch
// This has been tested with arduino UNO and Arduino Mega 2560
// your Mega 2560 may need an updated boot loader which fixes
// a crash when 3 exclaimation points appear consectively in a string
// and fixes its watchdog timer
// see http://www.desert-home.com/2012/05/arduino-mega2560-wrapping-up-bootloader.html
// see https://github.com/arduino/Arduino-stk500v2-bootloader/tree/master/goodHexFiles
**********************************************/
#define WATCHDOG_EN
#ifdef WATCHDOG_EN
  #include <avr/interrupt.h>    // for wathcdog ISR
  #include <avr/wdt.h>          // for the watchdog timer see bottom of setup routine
  
  uint8_t mcusr_mirror __attribute__ ((section (".noinit")));
  void get_mcusr(void) \
__attribute__((naked)) \
__attribute__((section(".init3")));

  void get_mcusr(void)
  {
      mcusr_mirror = MCUSR;
      MCUSR = 0;
      // We always need to make sure the WDT is disabled immediately after a
      //  reset, otherwise it will continue to operate with default values.
      //
      wdt_disable();
  }
#endif
//***************** end WATCHDOG_EN ***************

//*********** sleep requires WATCHDOG

#ifdef WATCHDOG_EN
  //#define SLEEP_EN    
  #ifdef SLEEP_EN
    // for Sleep
    // note from lowpowerlabs forum
    /*********************************
    The packet will only be received if the radio is listening.
    If it's sleeping it will not "hear" anything. And I am assuming it cant send anything too
    **********************************/
    boolean SLEEPING = false;
    #define USE_LOWPOWER_LIB
    #ifdef USE_LOWPOWER_LIB
      //get library from: https://github.com/LowPowerLab/LowPower
      //writeup here: http://www.rocketscream.com/blog/2011/07/04/lightweight-low-power-arduino-library/
      //#include <LowPower.h>
      /***************
      // even with USE_LOWPOWER_LIB undefined
      // the compiler report multiple definition of `__vector_6
      // aka WDT_vect so for now must comment out the include when USE_LOWPOER_LIB is not defined
      ****************/
    #else
      #include <avr/sleep.h>        // for sleep
      #include <avr/power.h>        // for low power while sleeping
      volatile int f_wdt=1;
      volatile int mywdt_ctr = 0;
    #endif
  #endif
#endif
//**************** end SLEEP_EN *****************


//************ for read vcc power
#define VCCMONITOR_EN
#ifdef VCCMONITOR_EN
  #if defined ARDURF1
    const float       LOW_BATTERY_VCC = 4.75;  // 4.75 volts
    const float       BATTERYDIVISOR  = 5.0;   // 5 volts
  #else
    #if defined ARDURF1S || defined PC2ARDURF1
      const float     LOW_BATTERY_VCC = 3.15;  // 3.15 volts
      const float     BATTERYDIVISOR  = 3.3;   // 3.3 volts
    #endif
  #endif
  float               lastEstimate2   = 0;     // previous result if filtering reading

  const int           ardurfVBattPin  = A7;     // Analog input pin that is connected the the battery monitor voltage
  const int           ardurfIChrgPin  = A6;     // Analog input pin that is connected the the charge current monitor voltage

  #define REPORTVCCSTATUSDELAY 60000  // Time in ms between updates
  unsigned long      previousVCCMillis   = 0;              // will store last time LED was updated
  long               vccstatusinterval    = REPORTVCCSTATUSDELAY;    // interval at which to blink (milliseconds)

 #endif
//************ end VCCMONITOR_EN *******************

void setup()
{
  #ifdef SERIAL_EN
    Serial.begin(SERIAL_BAUD);
    DEBUGflush();
  #endif

  /************************************
  // uncomment this just to see freeram when SERIAL_EN is not defined and none, some or all of the following defines are not defined:
  // RADIO_EN, WATCHDOG_EN, VCCMONITOR_EN
  #ifndef SERIAL_EN
    Serial.begin(SERIAL_BAUD);
    Serial.flush();
    Serial.print(F("Free Ram: "));
    Serial.println(freeRam()); 
  #endif
  ************************************/

  //Set up the watchdog timer. If it isn't reset every 4 or 8 seconds, the watchdog ISR will trigger
  #ifdef WATCHDOG_EN
    #ifdef SLEEP_EN
      #ifndef USE_LOWPOWER_LIB  // if using the lowpower lib, it configures its own watchdog
        WDT_Init_8();           // 8 seconds with ISR but without reboot
      #endif
    #else
      WDT_Init_4();           // 4 seconds with ISR and reboot
    #endif

    petthedog();
  #endif

  #ifdef SERIAL_EN
    showAppVersion();
  #endif
  
  #ifdef WATCHDOG_EN
    petthedog();
  #endif

  #ifdef RADIO_EN  
    #ifdef SERIAL_EN
      showHostArduRF();
    #endif
  #endif
  
  #ifdef WATCHDOG_EN
    petthedog();
  #endif

  showFreeSram();
  #ifdef WATCHDOG_EN
    petthedog();
  #endif

  #ifdef VCCMONITOR_EN
    showPwrSupplyVcc();  // check vcc
    showArduRFVBatt();
    showArduRFIChrg();
  #endif
  #ifdef WATCHDOG_EN
    petthedog();
  #endif

  #if defined RADIO_EN && defined USE_LOWPOWERLABS_RFM69_LIB
    #ifdef WATCHDOG_EN
      petthedog();
    #endif
    initLowPowerLabsRFM69();
    #ifdef WATCHDOG_EN
      petthedog();
    #endif
  #endif

  #if defined RADIO_EN && defined USE_RADIOHEAD_RF69_LIB
    #ifdef WATCHDOG_EN
      petthedog();
    #endif
    initRHmanager();
    #ifdef WATCHDOG_EN
    petthedog();
    #endif
  #endif

  #ifdef WATCHDOG_EN
    petthedog();
  #endif

  #ifdef SERIAL_EN
    #ifdef WATCHDOG_EN
      petthedog();
    #endif
    ShowMenu();
  #endif

  #ifdef WATCHDOG_EN
    petthedog();
  #endif
}

#if defined RADIO_EN && defined USE_LOWPOWERLABS_RFM69_LIB
void initLowPowerLabsRFM69()
{
  #ifdef SERIAL_EN
    showRadioLibraryUsed();
  #endif
  
  radio.initialize(FREQUENCY,MYNODEID,NETWORKID);
  DEBUGln(F("radio initialized OK"));
 
  // set modemconfig
  DEBUG(F("Setting ModemConfig to: "));
  DEBUGln(MODEMCONFIG);  // LowPowerLabs RFM69 init sets a default modem config
  
  // Turn on encryption
  DEBUGln(F("Setting Encryption Key On"));
  radio.encrypt(ENCRYPTKEY);
  encryptMode = true;

  // +13dBm, same as power-on default, +18dbm if IS_RFM69HW is defined
  DEBUG(F("Setting TxPower to: "));
  DEBUGln(TXPOWERSETTING);
  #ifdef IS_RFM69HW
    radio.setHighPower();  // it defaults to true
  #endif


  promiscuousMode = false;
  rxMode = true;    // allow listen if a LISTENER
  txMode = true;    // allow send if a SENDER
  
  #ifdef SERIAL_EN
    showRadioMode();
  #endif
}
#endif

#if defined RADIO_EN && defined USE_RADIOHEAD_RF69_LIB
void initRHmanager()
{
  radioPresent = true;  // assume its there
  #ifdef SERIAL_EN
    showRadioLibraryUsed();
  #endif

  // Defaults after init are 434.0MHz, modulation GFSK_Rb250Fd250, +13dbM, No Encryption, CRC On,
  // with 4 byte preamble and syncwords 0x2d, 0xd4 and not Promiscuous
  if (!manager.init())
  {
    DEBUGln(F("Radio init failed"));
    radioPresent = false;
    DEBUGln(F("No Radio"));
    return;
  }
  DEBUGln(F("radio initialized OK"));

  encryptMode = false;
  promiscuousMode = false;
  rxMode = true;    // allow listen if a LISTENER
  txMode = true;    // allow send if a SENDER

  snprintf_P(tempoutString1, (MAXARRAYSIZE - 1), PSTR("%s %s%2X %s%2X"), "Setting NetworkID to:", "0x", syncwords[0], "0x",syncwords[1]);
  DEBUGln(tempoutString1);
  driver.setSyncWords(syncwords, sizeof(syncwords));

  // GFSK_Rb250Fd250 (250Khz) is reasonably fast and reliable default speed and modulation
  // could try GFSK_Rb55555Fd50  slower (50Khz) but maybe better range see RH_RF69.h for other values
  // driver.MODEMCONFIG returns an index value from the  PROGMEM static const RH_RF69::ModemConfig MODEM_CONFIG_TABLE[]
  // see RH_69.cpp 
  // with MODEMCONGIG defined as GFSK_Rb250Fd250 it returns 18 as the index value 
  //snprintf_P(tempoutString1, (MAXARRAYSIZE - 1), PSTR("%s %d"), "Setting ModemConfig to:", RH_RF69::GFSK_Rb250Fd250);
  //DEBUGln(tempoutString1);
  DEBUG(F("Setting ModemConfig to: "));
  DEBUGln(F("GFSK_Rb250Fd250")); // need to change this text string if use a different config, see below

  //if (!driver.setModemConfig(driver.MODEMCONFIG))   // this does not work and actually locks up the radio, have to pull power to clear it
  if (!driver.setModemConfig(RH_RF69::GFSK_Rb250Fd250)) // if you change this need to change it in showRadioMode also
  {
    DEBUGln("setModemConfig failed");  
  }

  // 3 would be sufficient, but 4 is the same as RF22's
  DEBUG(F("Setting PreambleLength to: "));
  DEBUGln(PREAMBLELEN);
  driver.setPreambleLength(PREAMBLELEN);

  // change radio rxtx frequency
  char myBuf[8];
  snprintf_P(tempoutString1, (MAXARRAYSIZE - 1), PSTR("%s %s"), "Setting Frequency to:", ftoa(myBuf, FREQUENCY, 2));
  DEBUGln(tempoutString1);
  if (!driver.setFrequency(FREQUENCY))
  {
    DEBUGln("setFrequency failed");
  }

  // Turn on encryption
  DEBUGln(F("Setting Encryption Key On"));
  driver.setEncryptionKey(rh_ekey);
  encryptMode = true;

  // +13dBm, same as power-on default, +18dbm if IS_RFM69HW is defined
  //snprintf_P(tempoutString1, (MAXARRAYSIZE - 1), PSTR("%s %d), "Setting TxPower to:",TXPOWERSETTING);
  //DEBUGln(tempoutString1);
  DEBUG(F("Setting TxPower to: "));
  DEBUGln(TXPOWERSETTING);
  driver.setTxPower(TXPOWERSETTING);

  #ifdef WATCHDOG_EN
    petthedog();
  #endif
  
  #ifdef SERIAL_EN
    showRadioMode();
  #endif
  
  #ifdef WATCHDOG_EN
    petthedog();
  #endif

  // Defaults after init are 434.0MHz, modulation GFSK_Rb250Fd250, +13dbM, No Encryption, CRC On,
  // with 4 byte preamble and syncwords 0x2d, 0xd4 and not Promiscuous

  // After init the following can be changed later by the user if necessary.
  // Set up default configuration
  /**********************************
    uint8_t syncwords[] = { 0x2d, 0xd4 };     // syncwords act as a network id filter

    driver.setSyncWords(syncwords, sizeof(syncwords)); // Same as RF22's

    // Reasonably fast and reliable default speed and modulation
    // could try GFSK_Rb55555Fd50  slower but maybe better range
    // see RH_RF69.h for other values
    if (!driver.setModemConfig(RH_RF69::GFSK_Rb250Fd250))
    {
      DEBUGln("rh setModemConfig failed");
    }


    // 3 would be sufficient, but this is the same as RF22's
    driver.setPreambleLength(4);

    // radio frequency
    if (!driver.setFrequency(915.0))
    {
      DEBUGln("rh setFrequency failed");
    }


    // Turn off encryption
    driver.setEncryptionKey(NULL);
    encryptMode = false;

    // turn on promiscuousMode  -- this turns the radio into a packet sniffer, still to be tested
    driver.setPromiscuous(true);
    promiscuousMode = true;

    // with encryption
    driver.setEncryptionKey(rh_ekey);
    encryptMode = true;

    // +13dBm, same as power-on default
    driver.setTxPower(13);
  *********************************/

  // If you are using a high power RF69, you *must* set a Tx power in the
  // range 14 to 20 like this:
  // driver.setTxPower(14);
}
//******************************
#endif

#ifdef SERIAL_EN
void ShowMenu(void)
{
  #ifdef WATCHDOG_EN
    petthedog();
  #endif

  DEBUGln(F(""));
  DEBUGln(F("Menu"));
  DEBUGln(F("-----------"));
  DEBUGln(F("h = Show menu"));
  DEBUGln(F("t = Get rf69 temperature"));
  DEBUGln(F("B = Toggle Broadcast Mode on|off"));
  DEBUGln(F("D = Dump rf69 registers"));  // Uses Serial by default see rh_rf69.cpp
  DEBUGln(F("E = Toggle Encryption on|off"));
  DEBUGln(F("P = Toggle Promiscuous Mode on|off"));
  DEBUGln(F("R = Toggle Receiver on|off"));
  DEBUGln(F("S = Report Current Config"));
  DEBUGln(F("T = Toggle Transmitter on|off"));
  DEBUGln(F("? = Show menu"));
  DEBUGln(F(""));
  DEBUGln(F("Type command, followed by Enter\n"));

  #ifdef WATCHDOG_EN
    petthedog();
  #endif
}

void ProcessMenu()
{

  #ifdef WATCHDOG_EN
    petthedog();
  #endif
  if (Serial.available() > 0)
  {
    char incomingByte = Serial.read();
    switch (incomingByte)
    {
      case 'h':
      case '?':
        ShowMenu();
       break;

      // Ignore some characters, like the carriage return and line feed.
      case '\'':
      case '\r':
      case '\n':
        break;

      case 'D': //r=dump all register values
      {
        // both libs use Serial as a default to print out the data
        #ifdef USE_LOWPOWERLABS_RFM69_LIB
          radio.readAllRegs();
        #else
          #ifdef USE_RADIOHEAD_RF69_LIB
            driver.printRegisters();
          #endif
        #endif
        break;
      }

      case 'B': // toggle broadcastMode
      {
        broadcastMode = !broadcastMode;
        DEBUG(F("Broadcast Mode "));DEBUGln(broadcastMode ? "on" : "off");
        break;
      }
      
      case 'S': //report status
      {
        showStatus();
        break;
      }

      case 'E': //toggle encryption Mode
      {
        encryptMode = !encryptMode;
        if (encryptMode)
        {
          #ifdef USE_LOWPOWERLABS_RFM69_LIB
            radio.encrypt(ENCRYPTKEY);
          #else
            #ifdef USE_RADIOHEAD_RF69_LIB
              driver.setEncryptionKey(rh_ekey);
            #endif
          #endif
        }
        else
        {
          #ifdef USE_LOWPOWERLABS_RFM69_LIB
            radio.encrypt(null);
          #else
            #ifdef USE_RADIOHEAD_RF69_LIB
              driver.setEncryptionKey(NULL);
            #endif
          #endif
        }
        DEBUG(F("Encrypt Mode "));DEBUGln(encryptMode ? "on" : "off");
        break;
      }

      case 'P': // toggle promiscuousMode
      {
        promiscuousMode = !promiscuousMode;
        #ifdef USE_LOWPOWERLABS_RFM69_LIB
          radio.promiscuous(promiscuousMode);
        #else
          #ifdef USE_RADIOHEAD_RF69_LIB
            driver.setPromiscuous(promiscuousMode);  
          #endif
        #endif
        DEBUG(F("Promiscuous Mode "));DEBUGln(promiscuousMode ? "on" : "off");
        break;
      }

      case 'T': // toggle transmitter (aka listener) on off
      {
        txMode = !txMode;
        DEBUG(F("Transmitter "));DEBUGln(txMode ? "on" : "off");
        break;
      }
      case 'R': // toggle receiver (aka sender) on off
      {
        rxMode = !rxMode;
        DEBUG(F("Receiver "));DEBUGln(rxMode ? "on" : "off");
        break;
      }
      /*****************************/
      case 't':  // this has problems, get leading **** chars and it bricks the radio when using radiohead
      {
        showRadioTemperature();
        break;
      }
      /*****************************/

      default:
      {
        DEBUG(F("Unknown command: 0x"));
        DEBUGHexln(incomingByte,HEX);
        DEBUGln(F("Type '?' followed by Enter"));
        // Something could be wrong, delete incoming data in buffer
        while (Serial.available() > 0)
        {
          Serial.read();
          #ifdef WATCHDOG_EN
            petthedog();
          #endif
        }
        DEBUGln(F("-----------"));
        break;
    }
    }
    DEBUGln(F("Done."));
    DEBUGln(F("-----------"));
  }

  #ifdef WATCHDOG_EN
    petthedog();
  #endif
}

void showRadioTemperature()    // this has problems, get leading **** chars and it bricks the radio when using radiohead
{
  // Reads the on-chip temperature sensor.
  // The RF69 must be in Idle mode (= RF69 Standby) to measure temperature.
  // The measurement is uncalibrated and without calibration, you can expect it to be far from
  // correct.
  // returns the measured temperature, in degrees C from -40 to 85 (uncalibrated)
  
  // LowPowerLabs rfm69 lib calls setstandby mode inside the routine that reads the temperature see rfm69.lib
  // it issues a setMode(RF69_MODE_STANDBY); before reading the Temperature.
  
  // Radiohead rf69 lib does not call setstandby mode (aka setModeIdle() so you have to do it here see rh_69.cpp temperatureRead
  
  #ifdef USE_LOWPOWERLABS_RFM69_LIB
    byte temperature =  radio.readTemperature(-1); // -1 = user cal factor, adjust for correct ambient
    float fTemp = 1.8 * (byte)temperature + 32; // 9/5=1.8
  #else
    #ifdef USE_RADIOHEAD_RF69_LIB
      driver.setModeIdle();  // The RF69 must be in Idle mode (= RF69 Standby) to measure temperature.
      int8_t temperature = driver.temperatureRead();
      float fTemp = 1.8 * (int8_t)temperature + 32; // 9/5=1.8
    #endif
  #endif
  DEBUG(F("Radio Temp is: "));
  DEBUG(temperature);
  DEBUG(F("C, "));
  char myBuf[8];
  snprintf_P(tempoutString1, (MAXARRAYSIZE - 1), PSTR("%s%s"), ftoa(myBuf, fTemp, 2),"F");
  DEBUGln(tempoutString1); //converting to F loses some resolution, obvious when C is on edge between 2 values (ie 26C=78F, 27C=80F)
}

void showAppVersion()
{
  DEBUG(F("App Version: "));
  DEBUGln(AppVer);
}

#ifdef RADIO_EN
void showHostArduRF()
{
  #ifdef ARDURF1
    DEBUGln(F("Host ArduRF is: ARDURF1"));
  #endif
  #ifdef ARDURF1S
    DEBUGln(F("Host ArduRF is: ARDURF1S"));
  #endif
  #ifdef PC2ARDURF1
    DEBUGln(F("Host ArduRF is: PC2ARDURF1"));
  #endif
  #ifdef RADIO_SENDER
    DEBUGln(F("Radios primary mode is: RADIO_SENDER"));
  #endif
  #ifdef RADIO_LISTENER
    DEBUGln(F("Radios primary mode is: RADIO_LISTENER"));
  #endif
}
#endif

#ifdef RADIO_EN
void showRadioMode()
{
  #ifdef WATCHDOG_EN
    petthedog();
  #endif
      
  DEBUGln(F("\nRadio Configuration:"));

  #ifdef RADIO_SENDER
    #ifdef USE_LOWPOWERLABS_RFM69_LIB
      snprintf_P(tempoutString1, (MAXARRAYSIZE - 1), PSTR("%s %d %s"), "Sending at:", FREQUENCY==RF69_433MHZ ? 433 : FREQUENCY==RF69_868MHZ ? 868 : 915,  "Mhz");
      DEBUGln(tempoutString1);
      
      DEBUG(F("My NodeID is: "));
      DEBUGln(MYNODEID);
      DEBUG(F("My NetworkID is: "));
      DEBUGln(NETWORKID);
       
      DEBUG(F("Sending to NodeID:["));      
      DEBUG(SENDTONODEID);
      DEBUG(F("] once every ")); 
      DEBUG(SENDDELAY);
      DEBUGln(F(" milliseconds"));
      
      DEBUG(F("ModemConfig is: "));      
      DEBUGln(MODEMCONFIG);
      DEBUG(F("Tx Power is: "));
      DEBUGln(TXPOWERSETTING);
    #else
      #ifdef USE_RADIOHEAD_RF69_LIB
        snprintf_P(tempoutString1, (MAXARRAYSIZE - 1), PSTR("%s %d %s"), "Sending at", FREQUENCY==433.0 ? 433 : FREQUENCY==868.0 ? 868 : 915,  "Mhz");
        DEBUGln(tempoutString1);
          
        DEBUG(F("My NodeID is: "));
        DEBUGln(MYNODEID);      
        //DEBUG(F("My NetworkID is: "));        
        //DEBUGHex(syncwords[0],HEX);        
        //DEBUGHex(syncwords[1],HEX);
        //DEBUGln();  //send newline
        snprintf_P(tempoutString1, (MAXARRAYSIZE - 1), PSTR("%s %s%2X %s%2X"), "My NetworkID is:", "0x", syncwords[0], "0x",syncwords[1]);
        DEBUGln(tempoutString1);
        
        DEBUG(F("Sending to NodeID:["));      
        DEBUG(SENDTONODEID);
        DEBUG(F("] once every ")); 
        DEBUG(SENDDELAY);
        DEBUGln(F(" milliseconds"));
        
        // driver.MODEMCONFIG returns an index value from the  PROGMEM static const RH_RF69::ModemConfig MODEM_CONFIG_TABLE[]
        // see RH_69.cpp 
        // with MODEMCONGIG defined as GFSK_Rb250Fd250 it returns 18 as the index value
        DEBUGln(F("ModemConfig index val is: GFSK_Rb250Fd250")); // need to change this config string if use a different config, see initRhmanager

        DEBUG(F("Preamble Length is: "));
        DEBUGln(PREAMBLELEN);
        
        DEBUG(F("Tx Power is: "));
        DEBUGln(TXPOWERSETTING);
        
        DEBUGln(F("Client is Sending using rf69_reliable_datagram_server"));
      #endif
    #endif
  #endif
  
  #ifdef RADIO_LISTENER
    #ifdef USE_LOWPOWERLABS_RFM69_LIB
      snprintf_P(tempoutString1, (MAXARRAYSIZE - 1), PSTR("%s %d %s"), "Listening at:", FREQUENCY==RF69_433MHZ ? 433 : FREQUENCY==RF69_868MHZ ? 868 : 915,  "Mhz");
      DEBUGln(tempoutString1);
       
      DEBUG(F("My NodeID is: "));
      DEBUGln(MYNODEID);
      DEBUG(F("My NetworkID is: "));
      DEBUGln(NETWORKID);
      
      DEBUGln(F("Will reply to sender"));
      DEBUG(F("ModemConfig is: "));
      DEBUGln(MODEMCONFIG);
      DEBUG(F("Tx Power is: "));
      DEBUGln(TXPOWERSETTING);
    #else
      #ifdef USE_RADIOHEAD_RF69_LIB
        snprintf_P(tempoutString1, (MAXARRAYSIZE - 1), PSTR("%s %d %s"), "Listening at:", FREQUENCY==433.0 ? 433 : FREQUENCY==868.0 ? 868 : 915,  "Mhz");
        DEBUGln(tempoutString1);
       
        DEBUG(F("My NodeID is: "));
        DEBUGln(MYNODEID);
        
        snprintf_P(tempoutString1, (MAXARRAYSIZE - 1), PSTR("%s %s%2X %s%2X"), "My NetworkID is:", "0x", syncwords[0], "0x",syncwords[1]);
        DEBUGln(tempoutString1);
        
        DEBUGln(F("Will reply to sender"));
        
        // driver.MODEMCONFIG returns an index value from the  PROGMEM static const RH_RF69::ModemConfig MODEM_CONFIG_TABLE[]
        // see RH_69.cpp 
        // with MODEMCONGIG defined as GFSK_Rb250Fd250 it returns 18 as the index value
        // the below turns the radio into a brick causing a watchdog timeout, need a method to GET the current modem config
        //snprintf_P(tempoutString1, (MAXARRAYSIZE - 1), PSTR("%s %d"), "ModemConfig index is:", RH_RF69::GFSK_Rb250Fd250);
        //DEBUGln(tempoutString1);
        DEBUGln(F("ModemConfig index val is: GFSK_Rb250Fd250")); // need to change this config string if use a different config, see initRhmanager

        DEBUG(F("Preamble Length is: "));
        DEBUGln(PREAMBLELEN);
        
        DEBUG(F("Tx Power is: "));
        DEBUGln(TXPOWERSETTING);
        
        DEBUGln(F("Server is Listening using rf69_reliable_datagram_server"));
      #endif
    #endif
  #endif
 }
#endif

#ifdef RADIO_EN
void showRadioLibraryUsed()
{
  #ifdef USE_LOWPOWERLABS_RFM69_LIB
    DEBUGln(F("Radio Library is: LOWPOWERLABS_RFM69_LIB"));
  #endif
  #ifdef USE_RADIOHEAD_RF69_LIB
    DEBUGln(F("Radio Library is: RADIOHEAD_RF69_LIB"));
  #endif
}
#endif

void showStatus()
{
  showFreeSram();
  showAppVersion();
  #ifdef RADIO_EN
    showRadioLibraryUsed();
    showHostArduRF();
    showRadioMode();
    DEBUG(F("Broadcast Mode "));DEBUGln(broadcastMode ? "on" : "off");
    DEBUG(F("Encrypt mode "));DEBUGln(encryptMode ? "on" : "off");
    DEBUG(F("Promiscuous mode "));DEBUGln(promiscuousMode ? "on" : "off");
    DEBUG(F("Receiver "));DEBUGln(rxMode ? "on" : "off");
    DEBUG(F("Transmitter "));DEBUGln(txMode ? "on" : "off");
    //showRadioTemperature();  // this has problems, get leading **** chars and it bricks the radio when using radiohead
  #endif
  #ifdef VCCMONITOR_EN
    showPwrSupplyVcc();  // check vcc
    showArduRFVBatt();
    showArduRFIChrg();
  #endif
  #ifdef WATCHDOG_EN
    DEBUGln(F("Watchdog Timer enabled"));
  #else
    DEBUGln(F("Watchdog Timer disabled"));
  #endif
  #ifdef SLEEP_EN
    DEBUGln(F("Sleep Mode enabled"));
  #else
    DEBUGln(F("Sleep Mode disabled"));
  #endif
}
#endif  //****** end if SERIAL_EN

void loop()
{
  #ifdef SLEEP_EN
    #ifndef USE_LOWPOWER_LIB
      if(f_wdt == 1)   // if the watchdog barked do some work then go back to sleep
      {
        if (mywdt_ctr > 9)
        {
          mywdt_ctr = 0;
        }
        //delay(1000);
        DEBUG(F("Watchdog woke us up from sleep: "));  // geting high bit ascii chars in  begininng of string  on serial monitor
        DEBUGln(mywdt_ctr);
        mywdt_ctr++;
        // Don't forget to clear the flag.
        f_wdt = 0;
        delay(200);   // need a 50ms delay at 9600 baud before we enter sleep again so uart has time to complete its send
      }
    #endif
    SLEEPING = false;
    runnormalloopcode();
  #else
    runnormalloopcode();
  #endif // end of else for ifdef dosleep
}


void runnormalloopcode()
{
  #ifdef WATCHDOG_EN
    petthedog();
  #endif

  #ifdef SERIAL_EN
    ProcessMenu();
  #endif

  #ifdef VCCMONITOR_EN
    reportvccstatusDelay();
  #endif

  #if defined RADIO_EN // senders and listeners can listen
  if (rxMode)
  {
    #ifdef WATCHDOG_EN
      petthedog();
    #endif
    runRadioListen();
    #ifdef WATCHDOG_EN
      petthedog();
    #endif
  }

  if (txMode)
  {
    #ifdef WATCHDOG_EN
      petthedog();
    #endif
    sendDelay();
    #ifdef WATCHDOG_EN
      petthedog();
    #endif
  }
  #endif

  #ifdef SLEEP_EN
    // Re-enter sleep mode.
    enterSleep();  // uncomment this out if you want to sleep, sleep stops the millis timer
  #endif

  #ifdef WATCHDOG_EN
    petthedog();
  #endif

}

#if defined RADIO_EN
void runRadioListen()
{

  #ifdef WATCHDOG_EN
    petthedog();
  #endif

  //DEBUGln(F("Listening\n"));

  #ifdef USE_LOWPOWERLABS_RFM69_LIB
  if (radio.receiveDone())
  { 
    //DEBUGln(F("Rx ok\n"));
    
    #ifdef WATCHDOG_EN
      petthedog();
    #endif
    byte theSendersNodeID = radio.SENDERID;  // get the senders nodeid
    byte theTargetsNodeID = radio.TARGETID;  // get the target nodeid
    
    //Display a message on the serial monitor showing the senders NODEID
    DEBUG(F("Received from nodeid:["));
    DEBUG(theSendersNodeID);
    DEBUG(F("] "));

    if (theTargetsNodeID == MYNODEID || theTargetsNodeID == RF69_BROADCAST_ADDR)  //Check if the packets intended destination is this radio (GATEWAYID)
    {
      // if yes then process the data and send an Ack  

      //If Promiscuous mode is ON, show the NODEID the packet is intended for on the serial monitor
      if (promiscuousMode || theTargetsNodeID == RF69_BROADCAST_ADDR)
      {
        DEBUG(F("with to destination nodeid:["));
        DEBUG(theTargetsNodeID);
        DEBUG(F("] "));
      }

      //Add the received signal strength to the message to be displayed on the serial monitor
      DEBUG(F(" ....RX_RSSI["));
      DEBUG(radio.RSSI);
      DEBUGln(F("]"));
      
        
      //if (radio.DATALEN != sizeof(theData))    
      if (radio.DATALEN != sizeof(Payload))
      {
        DEBUGln(F("Invalid payload received, does not match Payload struct!"));
        return;
      }
      else
      {
        theData = *(Payload*)radio.DATA; //assume radio.DATA actually contains our struct and not something else
        DEBUG(F("Data is: nodeId="));
        DEBUG(theData.nodeId);
        DEBUG(F(" uptime="));
        DEBUG(theData.uptime);
        DEBUG(F(" vcc="));
        DEBUG(theData.vcc);
        DEBUG(F(" vbat="));
        DEBUG(theData.vbat);
        DEBUG(F(" ichrg="));
        DEBUG(theData.ichrg);
        DEBUG(F(" msg="));
        DEBUGln(theData.dataStr);  // may need to null terminate this to do string compares
      }

      //If an ACK (acknowledge) was requested by the data packets transmitters, send the ACK back to it and display "ACK sent"
      // on the serial monitor
      if (radio.ACK_REQUESTED)
      {
        #ifdef WATCHDOG_EN
          petthedog();
        #endif
        radio.sendACK();
        DEBUGln(F(" - ACK sent"));
      }
      DEBUGln();  // send newline
      #ifdef WATCHDOG_EN
        petthedog();
      #endif

      // send back a reply to the received message
      #ifdef RADIO_LISTENER
        if (theTargetsNodeID != RF69_BROADCAST_ADDR)
        {
          // send a packet NOT requesting an ACK back to the sender (if sender is not broadcasting)
          // This way both TX/RX MODE functions are tested
               
          DEBUG(F("Sending Reply to node:["));
          DEBUG(theSendersNodeID);
          DEBUG(F("] Reply is: "));
          getReplyToSend(theSendersNodeID);                
          DEBUG(theData.dataStr);  // may need to null terminate this to do string compares
          DEBUGln(); 
 
          if (sizeof(theData) > RF69_MAX_DATA_LEN )
          {
            DEBUG(F("Payload Length exceeds RF69_MAX_DATA_LEN: "));
            DEBUG(RF69_MAX_DATA_LEN);
            DEBUGln(F("reply cancelled"));
            return;
          }   
          DEBUG("Sending Payload struct with maximum of:(");
          DEBUG(sizeof(theData));
          DEBUGln(") bytes");          
          
          byte DESTINATION_NODEID = theSendersNodeID;
          DEBUG(F("Sending reply: "));DEBUG(theData.dataStr);DEBUG(F(" to nodeid:["));
          DEBUG(DESTINATION_NODEID); // send back a reply message to the sender
          DEBUGln(F("]\n"));
          radio.send(DESTINATION_NODEID, (const void*)(&theData), sizeof(theData),false);
        }
      #endif
      #ifdef WATCHDOG_EN
        petthedog();
      #endif
    }
    //If the destination nodeid is not this GATEWAYs NodeID and is not the broadcast nodeiid then show the Retries from the sender
    else
    {
      #ifdef WATCHDOG_EN
        petthedog();
      #endif
      // the radio.TARGETID != MYNODEID
      DEBUG(F("Received from nodeid:["));
      DEBUG(theSendersNodeID);
      DEBUG(F("] "));
      if (promiscuousMode)
      {
        DEBUG(F(" with to destination nodeid:["));
        DEBUG(theTargetsNodeID);
        DEBUG(F("]"));
      }
      
      DEBUG(F(" ....RX_RSSI["));
      DEBUG(radio.RSSI);
      DEBUGln(F("]"));

      #ifdef WATCHDOG_EN
        petthedog();
      #endif
          
      //if (radio.DATALEN != sizeof(Payload))      
      if (radio.DATALEN != sizeof(theData))
      {
        DEBUGln(F("Invalid payload received, not matching Payload struct!"));
        return;
      }
      else
      {
        // the retries from sender
        theData = *(Payload*)radio.DATA; //assume radio.DATA actually contains our struct and not something else
        DEBUG(F("Data is:  nodeId="));
        DEBUG(theData.nodeId);
        DEBUG(F(" uptime="));
        DEBUG(theData.uptime);
        DEBUG(F(" vcc="));
        DEBUG(theData.vcc);
        DEBUG(F(" vbat="));
        DEBUG(theData.vbat);
        DEBUG(F(" ichrg="));
        DEBUG(theData.ichrg);
        DEBUG(F(" msg="));
        DEBUGln(theData.dataStr);  // may need to null terminate this to do string compares
      }
      // tell the world the data was not for mynodeid
      DEBUGln(F("Data Not for me\n"));

      // do NOT check to see if ACK was requested
      // and do NOT send an ACK
      // since this data is not intented for this receiver
      // notice the lack of checking to see if an Ack was requested
      // and the lack of no code to send an Ack
      #ifdef WATCHDOG_EN
        petthedog();
      #endif
    }
  }
  #endif // end USE_LOWPOWERLABS_RFM69_LIB

  #ifdef USE_RADIOHEAD_RF69_LIB
    #ifdef WATCHDOG_EN
      petthedog();
    #endif
    if (radioPresent)
    {
      // radiohead io
      if (manager.available())   // start the receiver and if data available then process it
      { 
        //DEBUGln(F("Rx ok\n"));
        
        #ifdef WATCHDOG_EN
          petthedog();
        #endif
 
        // Wait for a message addressed to us from the client
        
        //uint8_t structlen = sizeof(Payload);     
        uint8_t structlen = sizeof(theData); 
        
        uint8_t fromnodeid;
        uint8_t tonodeid;
        
        if (manager.recvfromAck((uint8_t*)(&theData), &structlen, &fromnodeid, &tonodeid)) 
        {           
          if (tonodeid == MYNODEID || tonodeid == RH_BROADCAST_ADDRESS)  //Check if the packets intended destination is this radio (SERVER)
          {            
            DEBUG(F("Received from nodeid:["));
            DEBUG(fromnodeid);
            DEBUG(F("] "));
            // if yes then process the data and send an Ack  

            //If Promiscuous mode is ON, show the NODEID the packet is intended for on the serial monitor
            if (promiscuousMode || tonodeid == RH_BROADCAST_ADDRESS)
            {
              DEBUG(F("with to destination nodeid:["));
              DEBUG(tonodeid);
              DEBUG(F("] "));
            }
            //DEBUGln();  // send newline
 
            DEBUG(F(" ....RX_RSSI["));
            //DEBUG(driver.rssiRead());  // returns current rssi
            DEBUG(driver.lastRssi());    // returns last receoved packets rssi
            DEBUGln(F("]"));
              
            DEBUG(F("Data is: nodeId="));
            DEBUG(theData.nodeId);
            DEBUG(F(" uptime="));
            DEBUG(theData.uptime);
            DEBUG(F(" vcc="));
            DEBUG(theData.vcc);
            DEBUG(F(" vbat="));
            DEBUG(theData.vbat);
            DEBUG(F(" ichrg="));
            DEBUG(theData.ichrg);
            DEBUG(F(" msg="));
            DEBUGln(theData.dataStr);  // may need to null terminate this to do string compares 
            DEBUGln(); //send newline
            
            // send back a reply to the received message if sender was not broadcasting
            #ifdef RADIO_LISTENER
              if (tonodeid != RH_BROADCAST_ADDRESS)
              { 
                DEBUG(F("Sending Reply to node:["));
                DEBUG(fromnodeid);
                DEBUG(F("] Reply is: "));
                getReplyToSend(fromnodeid);                
                DEBUGln(theData.dataStr);  // may need to null terminate this to do string compares 
 
                if (sizeof(theData) > RH_RF69_MAX_MESSAGE_LEN )
                {
                  DEBUG(F("Payload Length exceeds RH_RF69_MAX_MESSAGE_LEN: "));
                  DEBUG(RH_RF69_MAX_MESSAGE_LEN);
                  DEBUGln(F("reply cancelled"));
                  return;
                }
                
                DEBUG("Sending Payload struct with maximum of:(");
                DEBUG(sizeof(theData));
                DEBUGln(") bytes");
                
                DEBUGln();
                // Send a reply back to the originating client
                //  server sends back to originating client: And hello back to you 
                if (!manager.sendtoWait((uint8_t*)(&theData), structlen, fromnodeid))
                {
                  DEBUGln(F("sendtoWait failed\n"));
                }
              }
            #endif
          }
          //If the destination nodeid is not this Servers NodeID or is not broadcast nodeid then show the Retries from the sender
          else
          {
            #ifdef WATCHDOG_EN
              petthedog();
            #endif  
            DEBUG(F("Received from nodeid:["));
            DEBUG(fromnodeid);
            DEBUG(F("] "));
            if (promiscuousMode)
            {
              DEBUG(F("with to destination nodeid:["));
              DEBUG(tonodeid);
              DEBUG(F("] "));
            }            
            DEBUG(F(" ....RX_RSSI["));
            //DEBUG(driver.rssiRead());  // return current rssi
            DEBUG(driver.lastRssi());    // returns last receoved packets rssi
            DEBUGln(F("]"));
       
            #ifdef WATCHDOG_EN
              petthedog();
            #endif
      
            // display the retries from sender            
            DEBUG(F("Data is: nodeId="));
            DEBUG(theData.nodeId);
            DEBUG(F(" uptime="));
            DEBUG(theData.uptime);
            DEBUG(F(" vcc="));
            DEBUG(theData.vcc);
            DEBUG(F(" vbat="));
            DEBUG(theData.vbat);
            DEBUG(F(" ichrg="));
            DEBUG(theData.ichrg);
            DEBUG(F(" msg="));
            DEBUGln(theData.dataStr);  // may need to null terminate this to do string compares
  
            // tell the world the data was not for mynodeid
            DEBUGln(F("Data Not for me\n"));
      
            // do NOT check to see if ACK was requested
            // and do NOT send an ACK
            // since this data is not intented for this receiver
            // notice the lack of checking to see if an Ack was requested
            // and the lack of no code to send an Ack
            #ifdef WATCHDOG_EN
              petthedog();
            #endif
          }
        }
        #ifdef WATCHDOG_EN
          petthedog();
        #endif
      }
    }
  #endif  // end USE_RADIOHEAD_RF69_LIB

  #ifdef WATCHDOG_EN
    petthedog();
  #endif
}
#endif

#if defined RADIO_EN
void getReplyToSend(uint8_t fromnodeid)
{     
  //fill in the struct with new values
  theData.nodeId = MYNODEID;
  theData.uptime = millis();
  //theData.temp = 66.45; //it's not to hot      
  theData.vcc   = getPwrSupplyVcc();     
  theData.vbat  = getArduRFVBatt();     
  theData.ichrg = getArduRFIChrg();
  snprintf_P(theData.dataStr, (MAXPAYLOADTEXTLEN), PSTR("%s"), "BAD DATA"); // keeps format string in Flash not ram
  // use below to send string, see radioListener
  if (fromnodeid == 2)
  {          
    snprintf_P(theData.dataStr, (MAXPAYLOADTEXTLEN), PSTR("%s"), "And hello back to you");           
  }
  else if (fromnodeid == 3)
  {
    snprintf_P(theData.dataStr, (MAXPAYLOADTEXTLEN), PSTR("%s"), "And good morning to you"); 
  }
  else if (fromnodeid == 4)
  {
    snprintf_P(theData.dataStr, (MAXPAYLOADTEXTLEN), PSTR("%s"), "No its not raining"); 
  }
  else if (fromnodeid == 1)
  {
    snprintf_P(theData.dataStr, (MAXPAYLOADTEXTLEN), PSTR("%s"), "Bad Data"); // if replying to self?
  }
  else
  {
    DEBUG(F(" getReplyToSend proc, unknown nodeid:["));
    DEBUG(fromnodeid);
    DEBUGln(F("]"));
    snprintf_P(theData.dataStr, (MAXPAYLOADTEXTLEN), PSTR("%s%d%s"), "unknown nodeid:[", fromnodeid, "]");
  }
}
#endif


#if defined RADIO_EN
void runRadioSend()
{
  #ifdef WATCHDOG_EN
    petthedog();
  #endif
  
  byte numRetries = 3;  // a 1 will actually retry 2 times see RFM69.cpp sendwithretry function

  #ifdef WATCHDOG_EN
    petthedog();
  #endif

  #ifdef USE_LOWPOWERLABS_RFM69_LIB  
    // using send struct
 
    if (sizeof(theData) > RF69_MAX_DATA_LEN )
    {
      DEBUG(F("Payload Length exceeds RF69_MAX_DATA_LEN: "));
      DEBUG(RF69_MAX_DATA_LEN);
      DEBUGln(F("send cancelled"));
      return;
    }
    
    uint8_t DESTINATION_NODEID = SENDTONODEID;
    toggleDestinationID = !toggleDestinationID;
    if (toggleDestinationID)
    {
      DESTINATION_NODEID = BOGUS_NODEID;
    }
    if (broadcastMode)
    {
      #ifdef WATCHDOG_EN
        petthedog();
      #endif
      DESTINATION_NODEID = RF69_BROADCAST_ADDR;
      DEBUG(F("Broadcasting: "));DEBUG(theData.dataStr);DEBUG(F(" to nodeid:[")); 
      DEBUG(DESTINATION_NODEID); // send to the good listener
      DEBUGln(F("] and NOT waiting for ACK..."));
      DEBUGln();  // send newline
      radio.send(DESTINATION_NODEID, (const void*)(&theData), sizeof(theData),false);
      #ifdef WATCHDOG_EN
        petthedog();
      #endif
    }
    else
    {
      #ifdef WATCHDOG_EN
        petthedog();
      #endif
      DEBUG(F("Sending: "));DEBUG(theData.dataStr);DEBUG(F(" to nodeid:["));
      DEBUG(DESTINATION_NODEID); // send to the good listener
      DEBUGln(F("] and waiting for ACK..."));
      if (radio.sendWithRetry(DESTINATION_NODEID, (const void*)(&theData), sizeof(theData),numRetries, ACK_TIME))
      {
        DEBUGln(F("ACK Received\n"));
      }
      else
      {
        DEBUGln(F("Send Failed\n"));
      }
      #ifdef WATCHDOG_EN
        petthedog();
      #endif
    }
    DEBUGln();
  #endif //********** USE_LOWPOWERLABS_RFM69_LIB

  #ifdef USE_RADIOHEAD_RF69_LIB
    #ifdef WATCHDOG_EN
      petthedog();
    #endif
    if (radioPresent)
    {
      if ((uint8_t)sizeof(theData) > RH_RF69_MAX_MESSAGE_LEN )
      {
        DEBUG(F("Payload Length exceeds RH_RF69_MAX_MESSAGE_LEN: "));
        DEBUG(RH_RF69_MAX_MESSAGE_LEN);
        DEBUGln(F("send cancelled"));
        return;
      }
      uint8_t DESTINATION_NODEID = SENDTONODEID;
      toggleDestinationID = !toggleDestinationID;
      if (toggleDestinationID)
      {
        DESTINATION_NODEID = BOGUS_NODEID;
      }
      if (broadcastMode)
      {
        DESTINATION_NODEID = RH_BROADCAST_ADDRESS;
        DEBUG(F("Broadcasting: "));DEBUG(theData.dataStr);DEBUG(F(" to nodeid:["));  
        DEBUG(DESTINATION_NODEID); // send to the good listener
        DEBUGln(F("] and NOT waiting for ACK..."));
        DEBUGln();  // send newline
      }
      else
      {
        // client sends a message to server
        // message is: Hello World
        DEBUG(F("Sending: "));DEBUG(theData.dataStr);DEBUG(F(" to nodeid:["));
        DEBUG(DESTINATION_NODEID);
        DEBUGln(F("] and waiting for ACK..."));
      }
 
      #ifdef WATCHDOG_EN
        petthedog();
      #endif
 
      //uint8_t structlen = sizeof(Payload);     
      uint8_t structlen = sizeof(theData);

      if (manager.sendtoWait((uint8_t*)(&theData), (uint8_t)structlen, (uint8_t)DESTINATION_NODEID))   
      {
        #ifdef WATCHDOG_EN
          petthedog();
        #endif
        if (!broadcastMode)  // if not broadcasting then check for ack
        { 
          // Now wait for a reply from the server
          uint8_t from;
          // server sends back a reply
          // reply is: And hello back to you 
          if (manager.recvfromAckTimeout((uint8_t*)(&theData), &structlen, 2000, &from))
          {
            #ifdef WATCHDOG_EN
              petthedog();
            #endif
            DEBUG(F("got reply from nodeid:["));
            DEBUG(from);
            DEBUG(F("], Data is: "));
            DEBUG(theData.dataStr);  // may need to null terminate this to do string compares
            
            DEBUG(F(" ....RSSI["));
            //DEBUG(driver.rssiRead());  // returns current rssi
            DEBUG(driver.lastRssi());    // returns last receoved packets rssi
            DEBUGln(F("]"));
            DEBUGln();
          }
          else
          {
            DEBUGln(F("No reply, is rf69_reliable_datagram_server running?\n"));
          }
        }
      }
      else
      {
        DEBUGln(F("sendtoWait failed\n"));
      }
 
      #ifdef WATCHDOG_EN
        petthedog();
      #endif
    }
    // end radiohead io
  #endif


  #ifdef WATCHDOG_EN
    petthedog();
  #endif
}
#endif


#if defined RADIO_EN
void sendDelay()
{
  // check to see if it's time to send data; that is, if the
  // difference between the current time and last time you sent
  // is bigger than the interval at which you want to
  // send.
  #ifdef WATCHDOG_EN
    petthedog();
  #endif
  unsigned long currentMillis = millis();
  unsigned long elapsedtime = currentMillis - previousSendMillis;
  #ifdef SLEEP_EN
      sendinterval = 1000;  // do not wait SENDDELAY seconds to send after waking up
  #endif

  if(elapsedtime > sendinterval)
  {
    #ifdef WATCHDOG_EN
      petthedog();
    #endif
    // save the last time you ran this
    previousSendMillis = currentMillis;
    {     
      //fill in the struct with new values
      theData.nodeId = MYNODEID;
      theData.uptime = millis();
      //theData.temp = 91.23; //it's hot!      
      theData.vcc   = getPwrSupplyVcc();      
      theData.vbat  = getArduRFVBatt();      
      theData.ichrg = getArduRFIChrg();
      snprintf_P(theData.dataStr, (MAXPAYLOADTEXTLEN), PSTR("%s"), "BAD DATA"); // keeps format string in Flash not ram
       
      // use below to send string, see radioListener
      if (MYNODEID == 2)
      {
        snprintf_P(theData.dataStr, (MAXPAYLOADTEXTLEN), PSTR("%s"), "Hello World");           
      }
      else if (MYNODEID == 3)
      {
        snprintf_P(theData.dataStr, (MAXPAYLOADTEXTLEN), PSTR("%s"), "Good Morning"); 
      }
      else if (MYNODEID == 4)
      {
        snprintf_P(theData.dataStr, (MAXPAYLOADTEXTLEN), PSTR("%s"), "Is it raining?"); 
      }
      else if (MYNODEID == 1)
      {
        snprintf_P(theData.dataStr, (MAXPAYLOADTEXTLEN), PSTR("%s"), "Listener is Sending"); 
      }
      else
      {
        DEBUG(F(" sendDelay proc, unknown nodeid:["));
        DEBUG(MYNODEID);
        DEBUGln(F("]"));
      }

      #ifdef VCCMONITOR_EN
        float myvolts = getPwrSupplyVcc();
        //showPwrSupplyVcc();
        if (myvolts < LOW_BATTERY_VCC)
        {
          snprintf_P(theData.dataStr, (MAXPAYLOADTEXTLEN), PSTR("%s"), "LOWBATTERY"); 
        }
      #endif
      if (MYNODEID == SENDTONODEID) // dont try to send to self
      {
        DEBUG(F("Cant send to your self, MyNodeID is["));// dont try to send to self
        DEBUG(MYNODEID);
        DEBUG(F("] ,"));
        DEBUG(F("SendToNodeID is["));
        DEBUG(SENDTONODEID);
        DEBUGln(F("]"));
      }
      else
      {
        DEBUG("About to Send Payload struct with a maximum of: (");
        DEBUG(sizeof(theData));
        DEBUGln(") bytes");
        DEBUG(F("Data is: "));
        DEBUG(F("nodeId="));
        DEBUG(theData.nodeId);
        DEBUG(F(" uptime="));
        DEBUG(theData.uptime);
        DEBUG(F(" vcc="));
        DEBUG(theData.vcc);
        DEBUG(F(" vbat="));
        DEBUG(theData.vbat);
        DEBUG(F(" ichrg="));
        DEBUG(theData.ichrg);
        DEBUG(F(" msg="));
        DEBUGln(theData.dataStr);  // may need to null terminate this to do string compares 
         
        #ifdef WATCHDOG_EN
          petthedog();
        #endif
        
        runRadioSend();
        
        #ifdef WATCHDOG_EN
          petthedog();
        #endif
        
      }
    }
  }
}
#endif

//*****************************************
int freeRam()
{
  extern int __heap_start, *__brkval;
  int v;
  return (int) &v - (__brkval == 0 ? (int) &__heap_start : (int) __brkval);
}


void showFreeSram()
{
  DEBUG(F("Free Ram: "));
  DEBUGln(freeRam());
}

//***********************************************
char *ftoa(char *a, double f, int precision)
{
  long p[] = {0,10,100,1000,10000,100000,1000000,10000000,100000000};

  char *ret = a;
  long heiltal = (long)f;
  itoa(heiltal, a, 10);
  while (*a != '\0') a++;
  *a++ = '.';
  long desimal = abs((long)((f - heiltal) * p[precision]));
  itoa(desimal, a, 10);
  return ret;
}

//************************************************************/
//*************************************

#ifdef WATCHDOG_EN
//initialize watchdog timer and its interrupt
// this works for 4 second watchdog reset and its ISR
void WDT_Init_4(void)
{
  //disable interrupts
  cli();
  //set up WDT interrupt
  WDTCSR = (1<<WDCE)|(1<<WDE);
  //Start watchdog timer with 4 second prescaller
  WDTCSR = (1<<WDIE)|(1<<WDE)|(1<<WDP3);          // this works for 4 seconds and sets up the ISR and a reset
  //Enable global interrupts
  sei();
}

// This works for 8 second watchdog, and its ISR, no reboot on watchdog timeout
void WDT_Init_8(void)
{
   // Setup the WDT
  cli();
  // Clear the reset flag.
  MCUSR &= ~(1<<WDRF);

  // In order to change WDE or the prescaler, we need to
  // set WDCE (This will allow updates for 4 clock cycles).
  //
  WDTCSR |= (1<<WDCE) | (1<<WDE);

  // set new watchdog timeout prescaler value
  //WDTCSR = 1<<WDP0 | 1<<WDP3; /* 8.0 seconds

  // set new watchdog timeout value to 1024K cycles (~8.0 sec)
  WDTCSR = (1<<WDP3) | (1<<WDP0);

  // Enable the WD interrupt only, do not enable a reset also
  WDTCSR |= _BV(WDIE);

  sei();
}

//Watchdog timeout ISR
// for use by WDINT_4 or WDINT_8 routines without sleep
#ifndef SLEEP_EN
  ISR(WDT_vect)
  {
    //Serial.print("Watchdog Interrupt routine running");
    //Serial.println();
    DEBUGln(F("WDT ISR Active, Rebooting"));
    DEBUGln();
  }
#endif


/****************************************************/
// watchdog ISR for Sleep
/***************************************************
 *  Name:        ISR(WDT_vect)
 *
 *  Returns:     Nothing.
 *
 *  Parameters:  None.
 *
 *  Description: Watchdog Interrupt Service. This
 *               is executed when watchdog times out.
 *
 ***************************************************/


#if defined SLEEP_EN && not defined USE_LOWPOWER_LIB
// even with USE_LOWPOWER_LIB undefined
// the compiler report multiple definition of `__vector_6
// aka WDT_vect so for now must use the LowPower lib to sleep
ISR(WDT_vect)
{
  if(f_wdt == 0)
  {
    f_wdt = 1;
  }
  else
  {
    DEBUGln(F("Sleep Watchdog Overrun!!")); // watch out, on the arduino mega 2560 R3 3 exclamtion points screws things up
  }
}
#endif

void petthedog()
{
  ; // noop
  #ifndef SLEEP_EN  // donot pet the dog if SLEEP MODE is enabled
    wdt_reset();  //Reset the watchdog
  #endif
}
#endif // end WATCHDOG_EN

/***************************************************
 *  Name:        enterSleep
 *
 *  Returns:     Nothing.
 *
 *  Parameters:  None.
 *
 *  Description: Enters the arduino into sleep mode and
 *    Sets the radio into low-power sleep mode. If successful, the radio will stay in sleep mode
 *    until woken up by changing mode it idle, transmit or receive (eg by calling send(), recv(), available() etc.)
 *    Caution: there is a time penalty as the radio takes a finite time to wake from sleep mode.
 *
 ***************************************************/
#ifdef SLEEP_EN
void enterSleep(void)
{
  DEBUGln(F("Time to go to sleep"));
  SLEEPING = true;

  #if defined RADIO_EN && defined USE_LOWPOWERLABS_RFM69_LIB
    radio.sleep(); // sleep the radio
  #endif

  #if defined RADIO_EN && defined USE_RADIOHEAD_RF69_LIB
    if (!driver.sleep()) // sleep the radio
    {
      DEBUGln(F("driver.sleep failed"));
    }
  #endif

  DEBUGflush();   // emptry the serial buffer
  #ifdef USE_LOWPOWER_LIB
    LowPower.powerDown(SLEEP_8S, ADC_OFF, BOD_OFF);
    // The program will continue from here after the WDT timeout
    delay(1000);                    // give system time to fully power up
    DEBUGln(F("Awake from sleep"));    
  #else
    set_sleep_mode(SLEEP_MODE_PWR_SAVE);    // could also use SLEEP_MODE_PWR_DOWN for lowest power consumption.
    //set_sleep_mode(SLEEP_MODE_PWR_DOWN);   
    sleep_enable();

    // Now enter sleep mode.
    sleep_mode();

    // The program will continue from here after the WDT timeout
    sleep_disable(); // First thing to do is disable sleep.

    // Re-enable the peripherals.
    power_all_enable();
    delay(1000);                    // give system time to fully power up
    DEBUGln(F("Awake from sleep"));
  #endif
}
#endif

/*****************************

******************************/

//There are some limitations in accuracy due to tolerances on the
//internal voltage reference. You can however, calibrate the scale
//factor for greater accuracy.
//This code runs on all Arduino variants as well as the ATtinyx4 series chips.

#ifdef VCCMONITOR_EN
void reportvccstatusDelay()
{
  // check to see if it's time to send data; that is, if the
  // difference between the current time and last time you sent
  // is bigger than the interval at which you want to send.
  unsigned long currentMillis = millis();

  if(currentMillis - previousVCCMillis > vccstatusinterval)
  {
    // save the last time you blinked the LED
    previousVCCMillis = currentMillis;
    {
      showArduRFVBatt();
      showArduRFIChrg();
      DEBUGln();
    }
  }
}

long readVcc() {
  // Read 1.1V reference against AVcc
  // set the reference to Vcc and the measurement to the internal 1.1V reference
  #if defined(__AVR_ATmega32U4__) || defined(__AVR_ATmega1280__) || defined(__AVR_ATmega2560__)
    ADMUX = _BV(REFS0) | _BV(MUX4) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);
  #elif defined (__AVR_ATtiny24__) || defined(__AVR_ATtiny44__) || defined(__AVR_ATtiny84__)
    ADMUX = _BV(MUX5) | _BV(MUX0);
  #elif defined (__AVR_ATtiny25__) || defined(__AVR_ATtiny45__) || defined(__AVR_ATtiny85__)
    ADMUX = _BV(MUX3) | _BV(MUX2);
  #else
    ADMUX = _BV(REFS0) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);
  #endif

  #if defined(__AVR_ATmega2560__)
    /****
    it took me a while to figure-out the problem, but on MEGA 2560, immediately after analogRead(A8), ADCL started returning zero. So every attempt to read from A8-A16 on Arduino MEGA will damage the functionality of readVcc().
    Iâ€™ve resolved the problem by adding:
    ADCSRB = 0;
    just before
    delay(2);
    **********************************/
    //ADCSRB &= ~_BV(MUX5); // Without this the function always returns -1 on the ATmega2560

    ADCSRB = 0;
  #endif

  delay(2); // Wait for Vref to settle
  ADCSRA |= _BV(ADSC); // Start conversion
  while (bit_is_set(ADCSRA,ADSC)); // measuring

  uint8_t low  = ADCL; // must read ADCL first - it then locks ADCH
  uint8_t high = ADCH; // unlocks both

  long result = (high<<8) | low;

  result = 1125300L / result; // Calculate Vcc (in mV); 1125300 = 1.1*1023*1000
  return result; // Vcc in millivolts
}
// filter the current result using a weighted average filter:
float filter(float rawValue, float weight, float lastValue)
{
  // run the filter:
  float result = weight * rawValue + (1.0-weight)*lastValue;
  // return the result:
  return result;
}

float getPwrSupplyVcc()
{
  float myvolts = float(readVcc()); // readVcc returns a long, volts is in millivolts, 5000 = 5 volts
  myvolts = myvolts/1000;
  return myvolts;
}
void showPwrSupplyVcc()
{
  //float myvolts = float(readVcc()); // readVcc returns a long, volts is in millivolts, 5000 = 5 volts
  // myvolts = myvolts/1000;
  float myvolts = getPwrSupplyVcc();

  // filter the sensor's result using weighted averaging:
  //float currentEstimate = filter(myvolts, 0.90, lastEstimate2);
  // lastEstimate2 = currentEstimate;
  //myvolts = currentEstimate;

  char myBuf[8];
  if (myvolts < LOW_BATTERY_VCC)
  {
    //sprintf(tempoutString1, "%s %s %s", "Vcc Low Bat:",ftoa(myBuf, myvolts, 2),"v");
    snprintf_P(tempoutString1, (MAXARRAYSIZE - 1), PSTR("%s %s %s"), "Vcc Low Bat:",ftoa(myBuf, myvolts, 2),"v");
  }
  else
  {
    //sprintf(tempoutString1, "%s %s %s", "Vcc is:",ftoa(myBuf, myvolts, 2),"v");
    snprintf_P(tempoutString1, (MAXARRAYSIZE - 1), PSTR("%s %s %s"), "Vcc is:",ftoa(myBuf, myvolts, 2),"v");
  }
  DEBUGln(tempoutString1);
}

//************************************************************
float getArduRFVBatt()
{
  int   sVal = analogRead(ardurfVBattPin);
  float fVal = calcArduRFBatt(sVal);
  return fVal;
}

//*****************************
int getArduRFIChrg()
{
  int sVal = analogRead(ardurfIChrgPin);  
  int sensorValue = map(sVal, 0, 375, 0, 400);    // map it to the range of the charge current in ma
  return sensorValue;
}

float calcArduRFBatt(int val)
{
  //float fdivisor = VCCDIVISOR;
  float fVal = float((val/1023) * BATTERYDIVISOR);
  return fVal;
}

void showArduRFVBatt()
{
  float fVal = getArduRFVBatt();
  char myBuf[8];
  snprintf_P(tempoutString1, (MAXARRAYSIZE - 1), PSTR("%s %s %s"), "ArduRF VBatt is:",ftoa(myBuf, fVal, 2),"v");
  DEBUGln(tempoutString1);
}

//*****************************
void showArduRFIChrg()
{
  int  sVal = getArduRFIChrg();  // read the voltage proportional to the battery charge current.
  snprintf_P(tempoutString1, (MAXARRAYSIZE - 1), PSTR("%s %d %s"), "ArduRF IChrg is:",sVal, "ma");
  DEBUGln(tempoutString1);
}
#endif  //*** end VCCMONITOR_EN


